
/****************************************************************/
/* FILE: dgaussfp.c                                             */
/* Two procedures for Gaussian elimination and forward/backward */
/* subtitution.							*/
/*      - fullPiv   LU-decomposition using full pivoting.	*/
/*      - fullBsv   Forward and backward substitution for the	*/
/*                  factored form produced by fullPiv.		*/
/****************************************************************/

#define max(A,B) ((A)>(B) ? (A) : (B) )
#define min(A,B) ((A)<(B) ? (A) : (B) )
#define sign(A) ( (A)>0?1:((A)<0?-1:0) )
#define abs(A) ( (A)>0?A:-A )

# include "dgaussfp.h"

/* Given that matrix A has been PLUQ factored by fullPiv, then
   use forward/backward substution algorithms to solve one of
   the following linear systems:
   Case:
      mode = 0  Given b, solve PLz = b for z.
           = 1  Given b, solve PLz = b for z, and then UQx = z for x.
           = 2  Given z, solve UQx = z for x.
   Note: x, z, and b must be distinct double arrays of length at least na.
   If x,z, or b does not appear in the above line for a given mode value,
   then the corresponding argument can be (double *) NULL. */
void fullBsv(double *a[], int na, int row[], int col[],
             double x[], double z[], double b[], int mode);

/* Perform LU-factorization with full pivoting on a na x na matrix A.
   Input:
     a: a[i] pointer to the i^{th} row of the matrix A
     na: dimension of A
     row[], col[]: uninitialized integer arrays of length at least na.
     eps: Machine epsilon, or larger (used to determine if the
          matrix is singular. 

   Output:
     a: overwritten with PLUQ factorization.  
     row, col:  the (i,j) element of L (or U) is stored in a[row[i]][col[j]]
                for i>j (or i<=j, respectively).  Here the diagonal elements
                of L are NOT stored, they are all equal to 1.
                In particular, the 1D array row encodes the effect of
                premultiplying by P, and the array col encodes the effect
                of postmultiplying by Q.
   Returns:
     0: successful factorization.
     i>0: if the matrix was first found to be nearly singluar at row i
         (i.e. all remaining pivots are less than or equal to 
          eps * max( abs( pivot).)  
          EG. for A = 0 (the na by na zero matrix), fullPiv returns 1...
          which means it was first discovered to be nearly singular while
          processing row number i = 1.
          EG. For A with a 1D null space, the returned i will typically be n-1
          (or smaller, in exceptional cases), with a small value for the
          very last pivot element.
 */
int fullPiv(double *a[], int na, int row[], int col[], double eps)
 {
  double piv,pivMax,pt,m,*pk,*pi,*pkj,*pij;
  int i,k,j,kp,jp,ci,cj;

  /* pivMax: Keeps the maximum absolute value of any pivot used so far. */
  pivMax = 0.0;
  /* Set up row, col permutation matrices */
  for(i=0; i<na; i++) row[i] = col[i] = i;
  /* Invariant: a[row[i]][col[j]] refers to the (i,j)-element of
     the permuted matrix.  Initially the permutation is
     the identity permutation. */

  /* Loop through rows */
  for(i=0; i<na; i++) {
   /* Search for next pivot element in the remaining block, 
      i.e. the bottom right block of the rearranged matrix a,
      starting at row i and column i */
   piv = -1.0;
   for( k=i; k<na; k++) {
    pk = a[row[k]];
    for( j=i; j<na; j++) {
     if ( (pt=abs(*(pk+col[j]))) > piv ) {
      kp=k; jp=j; piv=pt;
     }
    }
   }
   /* piv is now the maximum absolute value of the remaining element, 
      and this value occurs at row, col = (kp, jp) of the permuted
      matrix. */

   /* Check for nearly singular matrix */
   pivMax = max(piv,pivMax);
   if (piv<=eps*pivMax)
    return(i+1);  /* return the current row number at which singularity
                     was detected. */

   /* Perform interchange, swap rows kp and i, and swap cols jp and i.
      Note this swap is in terms of the current permutation, so we
      need only swap the values of row[kp] and row[i], and similarly
      swap the values of col[jp] and col[i]. */
   if (i!=kp) { k=row[kp]; row[kp]=row[i]; row[i]=k; }
   if (i!=jp) { k=col[jp]; col[jp]=col[i]; col[i]=k; }

   /* Eliminate: Use the ith row to introduce zeroes below the
      diagonal in the i^th column.  Write the multipliers into
      the lower triangular piece of the reduced matrix.  */
   ci=col[i]; pi=a[row[i]];
   for(k = i+1; k<na; k++) {
    pk = a[row[k]];
    pkj = pk+ci;
    m = *pkj = (*pkj)/(*(pi+ci)); /* The multiplier for row k */
    for(j = i+1; j<na; j++) {
     pkj=pk+col[j];
     *pkj = *pkj - m * ( *(pi+col[j]) );
    }
   }
  }
  return(0); /* Successful completion */
 }  /* End of fullPiv */

/* Given that matrix A has been PLUQ factored by fullPiv, then
   use forward/backward substution algorithms to solve one of
   the following linear systems:
   Case:
      mode = 0  Given b, solve PLz = b for z.
           = 1  Given b, solve PLz = b for z, and then UQx = z for x.
           = 2  Given z, solve UQx = z for x.
   Note: x, z, and b must be distinct double arrays of length at least na.
   If x,z, or b does not appear in the above line for a given mode value,
   then the corresponding argument can be (double *) NULL. */
void fullBsv(double *a[], int na, int row[], int col[],
             double *x, double *z, double *b, int mode)
 {
  int i,j,na1;
  double *pi,*pj,*ai;
  /* Forward solve PLz = b */
  if (mode<2) {
   for(i=0,pi=z; i<na; i++,pi++) {  /* recall L_{i,j} is stored in 
                                       A[row[i]][col[j]]. */
    *pi=b[row[i]]; ai=a[row[i]];
    for(j=0,pj=z; j<i; j++,pj++)
     *pi -= (*pj) * (*(ai+col[j]));
   }
  }
  /* Backward solve UQx = z */
  if (mode>0) {
   na1=na-1;
   for(i=na1; i>=0; i--) { /* recall U_{i,j} is stored in A[row[i]][col[j]] */
    pi=x+col[i]; ai=a[row[i]];
    *pi=z[i];
    for(j=na1; j>i; j--) {
     *pi -= (*(ai+col[j])) * x[col[j]];
    }
    *pi = *pi/(*(ai+col[i]));
   }
  }
 } /* End of fullBsv */

/* Compute the inverse of the na x na matrix a using
   a PLUQ factorization of a, and the forward/backward substitution
   to find each column of A-inverse.  Here a is overwritten by L and U,
   while row and col represent the permutations P and Q, respectively
   (see fullPiv, above).
   row and col must refer to integer arrays of size >= na.
   scr[0] and scr[1] must refer to scratch double arrays of size >= na.
   eps >= 0.0 is typically set to be the machine epsilon.
   The ith row of the inverse matrix (B = A^{-1}) is stored
   in the array referenced by b[i] (each of which must be of size >= na).
   Successful return: 0.
   Returned values >0 indicate that A is deemed to be singluar (depends
   on the given value of eps).  See fullPiv.
*/
int inverse(double *a[], double *b[], double *scr[],
            int row[], int col[], int na, 
            double eps)
 {
  double t;
  int j,k;

  /* Factor matrix a */
  if ( (j = fullPiv(a,na,row,col,eps)) > 0 )
   return(j);

  /* Backsolve for j^th column of inverse */
  for(j=0; j<na; j++) {
   for(k=0; k<na; k++) {
    if (k == j)
     *(scr[0]+k) = 1.0;
    else
     *(scr[0]+k) = 0.0;
   }
   
   /* Store the backsolved vector in the jth ROW of b. */
   fullBsv(a,na,row,col,b[j],scr[1],scr[0], (int) 1 );
  }

  /* Transpose b to form the inverse matrix. */
  for(j=0; j<na; j++)
   for(k=j+1; k<na; k++) {
    t = *(b[j]+k);
    *(b[j]+k) = *(b[k]+j);
    *(b[k]+j) = t;
   }

  return(0);

 } /* End of inverse */
