
/****************************************************************/
/* FILE: dgaussfp.h                                            */
/* Two procedures for Gaussian elimination and forward/backward */
/* subtitution.							*/
/*      - fullPiv   LU-decomposition using full pivoting.	*/
/*      - fullBsv   Forward and backward substitution for the	*/
/*                  factored form produced by fullPiv.		*/
/****************************************************************/

/* Perform LU-factorization with full pivoting on a na x na matrix A.
   Input:
     a: a[i] pointer to the i^{th} row of the matrix A
     na: dimension of A
     row[], col[]: uninitialized integer arrays of length at least na.
     eps: Machine epsilon, or larger (used to determine if the
          matrix is singular. 

   Output:
     a: overwritten with PLUQ factorization.  
     row, col:  the (i,j) element of L (or U) is stored in a[row[i]][col[j]]
                for i>j (or i<=j, respectively).  Here the diagonal elements
                of L are NOT stored, they are all equal to 1.
                In particular, the 1D array row encodes the effect of
                premultiplying by P, and the array col encodes the effect
                of postmultiplying by Q.
   Returns:
     0: successful factorization.
     i>0: if the matrix was first found to be nearly singluar at row i
         (i.e. all remaining pivots are less than or equal to 
          eps * max( abs( pivot).)  
          EG. for A = 0 (the na by na zero matrix), fullPiv returns 1...
          which means it was first discovered to be nearly singular while
          processing row number i = 1.
          EG. For A with a 1D null space, the returned i will typically be n-1
          (or smaller, in exceptional cases), with a small value for the
          very last pivot element.
 */
int fullPiv(double *a[], int na, int row[], int col[], double eps);

/* Given that matrix A has been PLUQ factored by fullPiv, then
   use forward/backward substution algorithms to solve one of
   the following linear systems:
   Case:
      mode = 0  Given b, solve PLz = b for z.
           = 1  Given b, solve PLz = b for z, and then UQx = z for x.
           = 2  Given z, solve UQx = z for x.
   Note: x, z, and b must be distinct double arrays of length at least na.
   If x,z, or b does not appear in the above line for a given mode value,
   then the corresponding argument can be (double *) NULL. */
void fullBsv(double *a[], int na, int row[], int col[],
             double x[], double z[], double b[], int mode);

/* Given that matrix A has been PLUQ factored by fullPiv, then
   use forward/backward substution algorithms to solve one of
   the following linear systems:
   Case:
      mode = 0  Given b, solve PLz = b for z.
           = 1  Given b, solve PLz = b for z, and then UQx = z for x.
           = 2  Given z, solve UQx = z for x.
   Note: x, z, and b must be distinct double arrays of length at least na.
   If x,z, or b does not appear in the above line for a given mode value,
   then the corresponding argument can be (double *) NULL. */
void fullBsv(double *a[], int na, int row[], int col[],
             double *x, double *z, double *b, int mode);
