/***
 FILE: mapping.v1.0.c   June 1999
 Utility routines for image warping. 

 Warp functions are used to define a new image as follows:
   newImage(x,y) = oldImage(warp(x,y; p))
 Here p = a vector of parameters.  The evaluation of oldImage
 between pixels is done by linear interpolation.

 WARNING:
  NO filtering is done during the image warping.  This should
  be included in order to control aliasing.  In particular, 
  the current warping code should not be used with warps involving
  significant (eg *2) scale changes in any direction.

 Notation:
  Images are represented by 1D (typically float) arrays, say image[ixy],
  where ixy = ix + iy * nx. Here (ix,iy) is the ix-th pixel along
  the iy-th scanline.
  Here (x,y) represents pixel coordinates, x to the left, y downwards,
  with the origin of an image taken at the top left corner.
  Image sizes are quoted as nx by ny, for nx pixels along each scan line
  and ny scan lines.
  In the comments, we use a simpler notation, namely image(x,y) or
  image(ix,iy) used to represent the image value at pixel location (x,y).
  If ix,iy are ints, then image(ix,iy) denotes image[ixy], with ixy as above.
  Otherwise, image(x,y) denotes an interpolated value.

 The basic warps considered are:
   affine:
    affineWarp(x_vec; p) = x_origin + a_vec + (I + A) * dx
   with dx = x_vec - x_origin.   Here x_origin is the arbitrary
   origin for the coordinate transform, typically chosen to be
   the center of the image region during the warp estimation process. 

   rational:
    rationalWarp(x_vec; p) = x_origin + (a + A * dx)/ (1+b^T * dx)
   with dx = x_vec - x_origin.

   radial distortion (quadratic):
     radialWarp(x_vec; p) = x_0 + a_vec + dx * (1 + a0 * ||dx||^2)
   with dx = x_vec - x_0.  Here x_0 is the center of the radial
   distortion (typically the piercing point of the optical axis).
  
 Important operations on a warp function include:
  - translateWarp
      newImage(x,y) = oldImage(translateWarp(x,y; p, x0, y0))
                    = oldImage(warp(x+x0, y+y0; p)) 
    This allows the origin (0,0) for newImage to be redefined, effectively
    changing the crop box.  This is redundant for groups which include
    translation, but is included here for other warps such as the radial
    warp.

  - inverseWarp 
      (x,y) = inverseWarp( warp(x,y; p), p) 
    this defines the reverse mapping
      oldImage(x,y) = newImage(inverseWarp(x,y; p))

  - composeWarps
      composeWarps(x,y; p, q) = warp( warp(x,y; p); q)
    This is used, for example, for warps defined over image sequences
    where the warp from the first frame to the last involves a composition
    of the warps between successive frames.

 Both affine and rational warps form a group, with well defined
 composition and inverse mappings which can be represented within
 the group.  That is,
   inverseWarp(x,y; p) = warp(x,y; inverseWarpCoeffs(p))
   composeWarps(x,y; p, q) = warp(x,y; composeWarpCoeffs(p, q))
 The group property is very useful during the modelling of successive
 warps over several frames of an image sequence.

 It is also useful to consider the special case of translations,
 which are simply a redefinition of the origin in image space:
    translateWarp(x,y; p, x0, y0) = warp(x,y; translateWarpCoeffs(p, x0, y0))
 Translation can be implemented for any warp family, independently of
 whether it forms a group or not.

 These operations are implemented by the routines below.  Composition is
 only implemented for affine and rational warps.  

 TTD:
  Add filtering during warping in order to control aliasing.
  
  Pass an image warp function to the warping routine(s), instead
  of using different routines for different types of warps.
***/

# include "macros.h"
# include "utils.h"
# include "mapping.h"

const char *mappingType[NUM_MAPPING_TYPES] =
               {"translation", "rotation", "affine", "rational"};

/**************** General interface routines ********************/

ImageWarpStruct * newImageWarpStruct()
{
  ImageWarpStruct * a;
  grabByteMemory((char **)&a, sizeof(ImageWarpStruct), "newImageWarpStructe");
  return(a);
}

void emptyImageWarpStruct(ImageWarpStruct *w)
{
  /* Place holder function: Nothing is currently allocated within an 
     ImageWarpStruct */
}

int getImageWarpType(ImageWarpStruct * a)
{
  int i;
  for(i=0; i<NUM_MAPPING_TYPES; i++) {
    if (strcmp(a->type, mappingType[i])==0)
      return(i);
    }
  return(INVALID_WARP_TYPE);
}

char const * getImageWarpName(int type)
{
  return((char const *) mappingType[type]);
}

int promptImageWarp(ImageWarpStruct *warp)
{
  char scrString[MAXLEN];
  int i, k;
  int isOk = 0;

  fprintf(stderr, " enter motion type, and nPar : ");
  readOneString(stdin, scrString);
  readOneInt(stdin, &(warp->nPar));
  for(i=0; i<NUM_MAPPING_TYPES; i++) {
    if (strcmp(scrString, mappingType[i])==0) {
      warp->type = mappingType[i];
      break;
    }
  }
  if (i==NUM_MAPPING_TYPES) 
    error("unrecognized mapping type", scrString);
 
  fprintf(stderr, " enter frozen[0..%d] : ", warp->nPar -1);
  warp->nFree=0;
  for(i=0; i<warp->nPar; i++) {
    readOneInt(stdin, &k);
    warp->frozen[i] = (k != 0);
    if (!warp->frozen[i]) warp->nFree++;
  }
  fprintf(stderr," enter warp params par[0..%d]: ", warp->nPar-1);
  for(i=0; i<warp->nPar; i++) {
    readOneFloat(stdin, &(warp->par[i]));
  }

  /* Check warp */
  if ((isOk=getImageWarpType(warp)) == INVALID_WARP_TYPE)
    fprintf(stderr,"Warning: promptImageWarp: Invalid input warp.");
  return(isOk);

} /* end: promptImageWarp */

/* Set argument to identity translation warp (if not a null pointer)
   Otherwise, when the argument warp pointer is NULL, a new WarpStruct
   is generated, initialized to the identity translation warp, and
   a pointer to it is returned */
ImageWarpStruct * setToIdentityWarp(ImageWarpStruct *warp, float *x_origin)
{
  if (warp == NULL) {
    warp = newImageWarpStruct();
  }
  warp->type = mappingType[TRANSLATION];
  warp->x_origin[0] = x_origin[0];
  warp->x_origin[1] = x_origin[1];
  warp->nPar = 2;
  warp->nFree = 2;
  warp->frozen[0] = warp->frozen[1] = FALSE;
  warp->par[0] = warp->par[1] = 0.0;
  return(warp);
}

/* Compute the warped image location of the pixel x[0-1], and
   return this location in xNew[0-1].
   Returns:
     INVALID_WARP_TYPE - if the warp is unrecognized.
     0 if successful
     Other (non-zero) warp specific flags, such as ON_HORIZ for 
     rational warps.
*/
int ptSampleImageWarp(float *x, ImageWarpStruct *warp, float *xNew)
{
  float dx[2];
  float ctheta, stheta, scl;
  int i, j, isOk;

  isOk = 0;
  switch ((warp->type)[0]) {

  case 't': /* Translation only */
    if (warp->nPar != 2)
      return(INVALID_WARP_TYPE);
    xNew[0] = x[0] + warp->par[0];
    xNew[1] = x[1] + warp->par[1];
    break;

  case 'a': /* Affine warp */
    if (warp->nPar != 6)
      return(INVALID_WARP_TYPE);
    isOk=affinePtSample(x, warp->x_origin, warp->par, xNew);
    break;

  case 'r':
    switch ((warp->type)[1]) {
    case 'o': /* Expect ROtation warp type, 2D translation and rotation */
      if (warp->nPar != 3) 
	return(INVALID_WARP_TYPE); /* Confused about the warp type */

      /*  See Eqn 1a below, with (I+A) = 2DRotMatrix(theta) */
      ctheta = cos(warp->par[2]);
      stheta = sin(warp->par[2]);
      dx[0] = x[0] - warp->x_origin[0];
      dx[1] = x[1] - warp->x_origin[1];
      xNew[0] = warp->x_origin[0] + warp->par[0]
	        + ctheta * dx[0] - stheta * dx[1];
      xNew[1] = warp->x_origin[1] + warp->par[1]  
	        + stheta * dx[0] + ctheta * dx[1];
      break;
    case 'a':       /*  Expect RAtional transformation */
      if (warp->nPar != 8) 
	return(INVALID_WARP_TYPE); /* Confused about the warp type */
      isOk=rationalPtSample(x, warp->x_origin, warp->par, xNew);
      break;
    default:
      return(INVALID_WARP_TYPE); /* Confused about the warp type */ 
    } /* completed switch on second character of warp type */
    break; /* completed outside case: first character of warp type is 'r' */  
    
  default:
    return(INVALID_WARP_TYPE); /* Confused about the warp type */
  } /* complete switch over first character of warp type */
      
  return(isOk);
}

/* Compute the jacobian of the warped image location with respect
   to the warp parameters, sampled at image position x,
   and return this 2 x (warp->nPar) matrix in jac.
   Returns:
     INVALID_WARP_TYPE - if the warp is unrecognized.
     0 if successful
     Other (non-zero) warp specific flags, such as ON_HORIZ for rational warps.
*/
int derivSampleImageWarp(float *x, ImageWarpStruct *warp,
			 float jac[2][MAX_MOTION_PAR])
{
  float xNew[2], dx[2];
  float ctheta, stheta, scl;
  int i, j, isOk;

  isOk = 0;

  switch ((warp->type)[0]) {

  case 't': /* Translation only */
    if (warp->nPar != 2)
      return(INVALID_WARP_TYPE);
    jac[0][1] = jac[1][0] = 0.0;
    jac[0][0] = jac[1][1] = 1.0;
    break;

  case 'a': /* Affine warp */
    if (warp->nPar != 6)
      return(INVALID_WARP_TYPE);
    isOk=affineJac(x, warp->x_origin, warp->par, jac);
    break;

  case 'r':
    switch ((warp->type)[1]) {
    case 'o': /* Expect ROtation warp type, 2D translation and rotation */
      if (warp->nPar != 3) 
	return(INVALID_WARP_TYPE); /* Confused about the warp type */

      /*  See Eqn 1a below, with (I+A) = 2DRotMatrix(theta) */
      ctheta = cos(warp->par[2]);
      stheta = sin(warp->par[2]);
      /* Translation params */
      jac[0][0] = jac[1][1] = 1.0;
      jac[0][1] = jac[1][0] = 0.0;
     
      /* Rotation param */
      dx[0] = x[0] - warp->x_origin[0];
      dx[1] = x[1] - warp->x_origin[1];
      jac[0][2] = -stheta * dx[0] - ctheta * dx[1];
      jac[1][2] = ctheta * dx[0] - stheta * dx[1];
      break;
    case 'a':       /*  Expect RAtional transformation */
      if (warp->nPar != 8) 
	return(INVALID_WARP_TYPE); /* Confused about the warp type */
      isOk=rationalJac(x, warp->x_origin, warp->par, jac);
      break;
    default:
      return(INVALID_WARP_TYPE); /* Confused about the warp type */ 
    } /* Completed switch on second character of warp type */
    break; /* complete case: first character of warp type is 'r' */  
    
  default:
    return(INVALID_WARP_TYPE); /* Confused about the warp type */
  } /* complete switch over first character of warp type */
      
  return(isOk);
}

/* Given frame sample times t-baseDelay, and target times
      t-delay[0], t-delay[1], ...  t-delay[nTarget-1]
    for 0<=delay[k]<=maxDelay.
   Given the constant forward warp params, fwd, between each pair
   of frames t-k to t-(k-1), k<=maxDelay<=1
  OUTPUT:
    Compute the cummulative warps, 
       *cummWarp[k],  k=0,...,nTarget-1,
    from the base frame to the target frame delay[k], k=0,...,nTarget-1.

  EG: baseFrame = 0, nTarget = 3, maxDelay = 3;
         delay[k] = k+1,
    Then frames t-0,...,t-3 are considered. fwdWarp needs to be the
    warp from frame t-(k+1) to t-k, for k=0,1,2,3.
    On output cummWarp[k], k=0,..,2 gives the backwards
    warp from frame t-0 to target frame t-k, k = 1,..,3. 

  EG: baseFrame = 1, delay[*] = ( 0 2), nTarget = 2, maxDelay = 2 then 
    frame delay 1 is the base frame and cummWarp[0] maps this baseframe
    forward to the frame at time 0, and cummWarp[1] maps baseframe 1
    backwards to target frame 2. **/
void setCompositeWarps(ImageWarpStruct *cummWarp[], 
		      int baseDelay, int *delay, int nTarget,
		      ImageWarpStruct *fwdWarp, int maxDelay)
{
  ImageWarpStruct *currentWarp, *tmpWarp, *bkwdWarp;
  int i, iTarget, iFuture;
  int currentDelay, targetDelay, saveDelay;
  int *iSort;

  if (maxDelay <=0 ) 
    error("setRelativeWarps: maxDelay is negative or zero.","");

  /* Sort the delays */
  grabIntMemory(&iSort, nTarget, "setRelativeWarps: iSort");
  {
    float *fDelay;
    grabFloatMemory(&fDelay, nTarget, "setRelativeWarps: fDelay");
    
    for(i=0; i<nTarget; i++)
      fDelay[i] = -(float) delay[i];   /* Note the sign reversal so
                                          the subsequent sort gives
				          delays from smallest to largest */
    sortFloatArray(iSort, fDelay, nTarget);  /* Sorts in decreasing order */
    free(fDelay);
  }
  /* Now delay[iSort[k]], k=0, ... nTarget-1 is sorted smallest to largest */

  /* Find the largest delay that is still in the future of the baseFrame 
     (if any) */
  iFuture = -1;
  for(i=0; i<nTarget; i++)
    /* Invariant: delay[iSort[0..iFuture] <= baseDelay */
    if (delay[iSort[i]] > baseDelay)
      break;
    else
      iFuture = i;

  if (iFuture >= 0) { /* Note: delay[iSort[0..iFuture] <= baseDelay */
    /* Need to warp the base frame forward to delays delay[iSort[k]],
       for k=0..iFuture.  Note these delays are sorted by iSort[k], and
       it is useful to compose the warps in the order of increasing time
       from the base frame (decreasing delay):
       delay[iSort[iFuture]] ... delay[iSort[0]]  */

    /* Build the identity warp having the same origin as the base
       frame.  This identity warp is initially set to be a translation,
       it will be extended to other types of warps, if necessary, when
       it is composed with other warps. */
    currentWarp = setToIdentityWarp((ImageWarpStruct *) NULL, 
		      fwdWarp->x_origin);
    currentWarp->frozen[0] = fwdWarp->frozen[0];
    currentWarp->frozen[1] = fwdWarp->frozen[1];
    currentDelay = baseDelay;
    saveDelay = baseDelay+1; /* saveDelay indicates delays we wish to
			        save the cummulative warps at.  Here
			        we set it to a dummy initial value. */
    /* Compose incremental warps forward one frame using 
       the constant forward warp, fwdWarp */
    for(i=iFuture; i>=0; i--) {
      targetDelay = delay[iSort[i]];
      while(currentDelay > targetDelay) {  /* move to the future */
	if (currentDelay < 1)
	  error("setRelativeWarps was given a negative frame delay","");
	if (composeImageWarps(currentWarp, fwdWarp, &tmpWarp) != 0)
	  error("can't compose an incremental forward warp","");
        if (currentDelay < saveDelay) {
	  /* Don't need to save the currentWarp */
	  emptyImageWarpStruct(currentWarp); free(currentWarp);
        }
        currentWarp = tmpWarp;
        currentDelay--;
      }
      /* Here targetDelay == currentDelay == delay[iSort[i]] */
      cummWarp[iSort[i]] = currentWarp;
      saveDelay = currentDelay;  /* Don't delete this sucker next iteration */
    }
  }  
  /* Completed the warps to all frames in the future of base frame. */

  if (iFuture < nTarget-1) { 
    /* Note: delay[iSort[iFuture+1..nTarget-1] > baseDelay */
    /* Need to warp the base frame backwards to delays delay[iSort[k]],
       k=iFuture .. nTarget-1.  Note these delays are sorted by iSort[k], and
       it is useful to compose the warps in the order of increasing DELAY:
       delay[iSort[iFuture]] ... delay[iSort[nTarget-1]]  */

    /* Build the identity warp having the same origin as the base
       frame.  This identity warp is initially set to be a translation,
       it will be extended to other types of warps, if necessary, when
       it is composed with other warps. */
    currentWarp = setToIdentityWarp((ImageWarpStruct *) NULL, 
		      fwdWarp->x_origin);
    currentWarp->frozen[0] = fwdWarp->frozen[0];
    currentWarp->frozen[1] = fwdWarp->frozen[1];
    currentDelay = baseDelay;
    saveDelay = baseDelay-1;  /* don't wish to save the current warp */
    /* Compose incremental warps backward one frame using the
       inverse of the constant forward warp */
    if (invertImageWarp(fwdWarp, &bkwdWarp) != 0) 
      error("can't invert incremental warp","");
    for(i=iFuture+1; i<nTarget-1; i++) {
      targetDelay = delay[iSort[i]];
      while(currentDelay < targetDelay) {
	if (currentDelay > maxDelay-1)
	  error("setCompositeWarps was given an invalid frame delay,",
                "larger than maxDelay-1");
	/* Compose incremental warps backward one frame */
	if (composeImageWarps(currentWarp, bkwdWarp, &tmpWarp)
	       != 0)
	  error("can't compose an incremental backward warp","");
        if (currentDelay > saveDelay) {
	  /* Don't need to save the currentWarp */
	  emptyImageWarpStruct(currentWarp); free(currentWarp);
        }
        currentWarp = tmpWarp;
        currentDelay++;
      }
      /* Here targetDelay == currentDelay == delay[iSort[i]] */
      cummWarp[iSort[i]] = currentWarp;
      saveDelay = currentDelay;
    }
    emptyImageWarpStruct(bkwdWarp); free(bkwdWarp);
  }  
  /* Completed the warps to all frames older than the base frame. */
  free(iSort);
} /* End: setRelativeWarps */

/* Compose warps represented by warp structs.
   The composed warp is returned in *pResult,
        result(x) = second( first (x))
   Returns 0 if successful.
*/
int composeImageWarps(ImageWarpStruct *first, ImageWarpStruct *second, 
		      ImageWarpStruct **pResult)
{
  ImageWarpStruct *a, *b, *r;
  float ctheta, stheta, dx[2];
  float aPar[8], bPar[8];
  int type1, type2, typeResult;
  int fz, k;
  int isOk = 0;

  /* Check the types of first and second */
  type1 = getImageWarpType(first);
  type2 = getImageWarpType(second);
  if (type1 < 0 || type2 < 0)
    return(INVALID_WARP_TYPE);
  typeResult = max(type1, type2); /* BEWARE: This is somewhat limited.  If
         more warps are included then something more complicated
         will need to be done. */

  if (type1 < typeResult) {
    a = newImageWarpStruct();
    *a = *first;
    convertImageWarpType( a, typeResult);
  } else
    a = first;
  if (type2 < typeResult) {
    b = newImageWarpStruct();
    *b = *second;
    convertImageWarpType( b, typeResult);
  } else
    b = second;
 
  /* Allocate the result */
  r = *pResult = newImageWarpStruct();
  switch(typeResult) {  /* Allocate and set **pResult */
  case TRANSLATION: 
    r->type = mappingType[TRANSLATION];
    r->x_origin[0] = a->x_origin[0];
    r->x_origin[1] = a->x_origin[1];
    r->par[0] = a->par[0] + b->par[0];
    r->par[1] = a->par[1] + b->par[1];
    r->nPar = 2;
    r->frozen[0] = a->frozen[0] && b->frozen[0];
    r->frozen[1] = a->frozen[1] && b->frozen[1];
    r->nFree = 0;
    if (!r->frozen[0]) r->nFree++;
    if (!r->frozen[1]) r->nFree++;
    break;

  case ROTATION:
    /* xa = Ra (x - xa0) + xa0 + ta
       xr = Rb (xa - xb0) + xb0 + tb
          = Rb Ra (x-xa0) + Rb( xa0 + ta) -Rb xb0 + xb0 + tb
          = Rb Ra (x-xa0) + xa0 + [Rb(xa0-xb0) - (xa0 - xb0)] + Rb ta + tb
          = Rb Ra (x-xa0) + xa0 + [Rb-I](xa0-xb0) + Rb ta + tb   */
    r->type = mappingType[ROTATION];
    r->nPar = 3;
    r->x_origin[0] = a->x_origin[0];
    r->x_origin[1] = a->x_origin[1];
    ctheta = cos(b->par[2]);
    stheta = sin(b->par[2]);
    r->par[0] = b->par[0] + ctheta * a->par[0] - stheta * a->par[1];
    r->par[1] = b->par[1] + stheta * a->par[0] + ctheta * a->par[1];
    dx[0] = a->x_origin[0] - b->x_origin[0];
    dx[1] = a->x_origin[1] - b->x_origin[1];
    r->par[0] += (ctheta - 1.0) * dx[0] - stheta * dx[1];
    r->par[1] += stheta * dx[0] + (ctheta - 1.0) * dx[1];
    r->par[2] = a->par[2] + b->par[2];
    /* Treat frozen cases:
        - all frozen
        - both translations frozen, same center => axial rotation.
	- both rotations frozen => translation only */
    r->frozen[0] = r->frozen[1] = r->frozen[2] = FALSE;
    if (a->frozen[0] && a->frozen[1] &&        
        b->frozen[0] && b->frozen[1]) {
      if (a->frozen[2] && b->frozen[2]) { /* All frozen */
	r->frozen[0] = r->frozen[1] = r->frozen[2] = TRUE;
      } else if ((abs(dx[0])<MACHINE_EPS) && (abs(dx[1])<MACHINE_EPS)) {
        /* Axial rotation */
	r->frozen[0] = r->frozen[1] = TRUE;
	r->frozen[2] = FALSE;
      }  /* Otherwise, rotation with frozen offset... 
	    generically nothing is be frozen */
    } else {
      if (a->frozen[2] && b->frozen[2]) 
	r->frozen[2] = TRUE;
    }
    r->nFree = 0;
    if (!r->frozen[0]) r->nFree++;
    if (!r->frozen[1]) r->nFree++;
    if (!r->frozen[2]) r->nFree++;
    break;

  case AFFINE:
    r->type = mappingType[AFFINE];
    r->x_origin[0] = a->x_origin[0];
    r->x_origin[1] = a->x_origin[1];
    for(k=0; k<6; k++) r->par[k] = a->par[k];
    composeAffine(r->x_origin, r->par, b->x_origin, b->par);
    r->nPar = 6;
    /* Treat frozen cases:
       - all frozen
       - affine terms frozen
       - axial distortion (translation frozen) */
    for(k=0; k<6; k++) r->frozen[k] = FALSE;
    fz = TRUE;
    for(k=2; k<6; k++)
      if (!a->frozen[k] || !b->frozen[k]) {
	fz = FALSE;
	break;
      }
    if (fz) {  /* Affine terms (at least) all frozen */
      for(k=2; k<6; k++) r->frozen[k] = TRUE;
      
      for(k=0; k<2; k++)
	if (!a->frozen[k] || !b->frozen[k]) {
	  fz = FALSE;
	  break;
	}
      if (fz)  /* All frozen */
	for(k=0; k<2; k++)  r->frozen[k] = TRUE;

    } else {  /* Affine terms not frozen, check for radial distortion */
      dx[0] = a->x_origin[0] - b->x_origin[0];
      dx[1] = a->x_origin[1] - b->x_origin[1];
      if (abs(dx[0]) < MACHINE_EPS && abs(dx[1]) < MACHINE_EPS) {
	fz = TRUE; /* same origin, check no translation */
	for(k=0; k<2; k++)
	  if (!a->frozen[k] || !b->frozen[k]) {
	    fz = FALSE;
	    break;
	  }
	if (fz)  /* Axial distortion ok */
	  for(k=0; k<2; k++)  r->frozen[k] = TRUE;
      }
    }
    r->nFree = 0;
    for(k=0; k<6; k++) if (!r->frozen[k]) r->nFree++;
    break;
    
  case RATIONAL:
    r->type = mappingType[RATIONAL];
    r->x_origin[0] = a->x_origin[0];
    r->x_origin[1] = a->x_origin[1];
    for(k=0; k<8; k++) r->par[k] = a->par[k];
    composeRational(r->x_origin, r->par, b->x_origin, b->par);
    r->nPar = 8;
    /* Treat frozen cases:
       - all frozen
    */
    for(k=0; k<8; k++) r->frozen[k] = FALSE;
    fz = TRUE;
    for(k=0; k<8; k++)
      if (!a->frozen[k] || !b->frozen[k]) {
	fz = FALSE;
	break;
      }
    if (fz)  /* all frozen */ 
      for(k=0; k<8; k++) r->frozen[k] = TRUE;
      
    r->nFree = 0;
    for(k=0; k<8; k++) if (!r->frozen[k]) r->nFree++;
    break;

  default:
    isOk = INVALID_WARP_TYPE;
  }
  if (a != first) {
    emptyImageWarpStruct(a);
    free(a);
  }
  if (b != second) {
    emptyImageWarpStruct(second);
    free(b);
  }
  return(isOk);
}

int moveImageWarpOrigin(ImageWarpStruct *a, float *newOrig, 
			ImageWarpStruct **pResult)
{
  ImageWarpStruct *r;
  float ctheta, stheta, dx[2];
  float aPar[8], bPar[8];
  int type;
  int fz, k;
  int isOk = 0;

  /* Check the types of first and second */
  type = getImageWarpType(a);
 
  /* Allocate the result */
  r = *pResult = newImageWarpStruct();
   
  switch(type) {  /* Set the inverse warp **pResult */
  case TRANSLATION:
    *r = *a; 
    r->x_origin[0] = newOrig[0];
    r->x_origin[1] = newOrig[1];
    break;

  case ROTATION:
    /* xa = Ra (x - xa0) + xa0 + ta
       xr0 = newOrig
       xr = R(a)(x - xr0) + xr0 + R(a) (xr0- xa0) + xa0 - xr0 + ta
          = R(a)(x - xr0) + xr0 + ptSampleWarp(xr0, a, dx) - xr0
    */
    *r = *a;
    r->x_origin[0] = newOrig[0];
    r->x_origin[1] = newOrig[1];
    ptSampleImageWarp(newOrig, a, dx);
    r->par[0] = dx[0] - r->x_origin[0];
    r->par[1] = dx[1] - r->x_origin[1];
    /* Treat frozen cases:
        - all frozen
	- rotations frozen => translation only */
    r->frozen[0] = r->frozen[1] = r->frozen[2] = FALSE;
    if (a->frozen[0] && a->frozen[1]) {     
      if (a->frozen[2]) { /* All frozen */
	r->frozen[0] = r->frozen[1] = r->frozen[2] = TRUE;
      } 
    } else {
      if (a->frozen[2]) {
	r->frozen[2] = TRUE;
      }
    }
    break;

  case AFFINE:
    *r = *a;
    for(k=0; k<6; k++) r->par[k] = a->par[k];
    r->x_origin[0] = newOrig[0];
    r->x_origin[1] = newOrig[1];
    /* Move center to forward warped center position */
    moveAffineOrigin(a->x_origin, r->par, r->x_origin);
    /* Treat frozen cases:
       - all frozen
       - affine terms frozen
     */
    for(k=0; k<6; k++) r->frozen[k] = FALSE;
    fz = TRUE;
    for(k=2; k<6; k++)
      if (!a->frozen[k]) {
	fz = FALSE;
	break;
      }
    if (fz) {  /* Affine terms (at least) all frozen */
      for(k=2; k<6; k++) r->frozen[k] = TRUE;
      
      for(k=0; k<2; k++)
	if (!a->frozen[k]) {
	  fz = FALSE;
	  break;
	}
      if (fz)  /* All frozen */
	for(k=0; k<2; k++)  r->frozen[k] = TRUE;

    }
    break;
    
  case RATIONAL:
    *r = *a;
    r->x_origin[0] = newOrig[0];
    r->x_origin[1] = newOrig[1];
    moveRationalOrigin(a->x_origin, r->par, r->x_origin);
    /* Treat frozen case:
       - all frozen
    */
    fz = TRUE;
    for(k=0; k<8; k++)
      if (!a->frozen[k]) {
	fz = FALSE;
	break;
      }
    if (fz)  /* all frozen */ 
      for(k=0; k<8; k++) r->frozen[k] = TRUE;
    else
      for(k=0; k<8; k++) r->frozen[k] = FALSE;      
    break;

  default:
    isOk = INVALID_WARP_TYPE;
  } /* end of switch over warp types */

  r->nFree = 0;  
  for(k=0; k<r->nPar; k++)
    if (!r->frozen[k]) r->nFree++;
  return(isOk);
 
}  /** end moveImageWarpOrigin */

int invertImageWarp(ImageWarpStruct *a, ImageWarpStruct **pResult)
{
  ImageWarpStruct *r;
  float ctheta, stheta, dx[2];
  float aPar[8], bPar[8];
  int type;
  int fz, k;
  int isOk = 0;

  /* Check the types of first and second */
  type = getImageWarpType(a);
 
  /* Allocate the result */
  r = *pResult = newImageWarpStruct();
  /* Set the origin of the inverse to equal the forward warp
     of the forward map's origin. */
  if (ptSampleImageWarp(a->x_origin, a, r->x_origin) != 0) {
    fprintf(stderr, "invertImageWarp: Cannot sample warp at origin\n");
    return(-1);
  }
   
  switch(type) {  /* Set the inverse warp **pResult */
  case TRANSLATION: 
    r->type = mappingType[TRANSLATION];
    r->par[0] = -a->par[0];
    r->par[1] = -a->par[1];
    r->nPar = 2;
    r->frozen[0] = a->frozen[0];
    r->frozen[1] = a->frozen[1];
    r->nFree = 0;
    break;

  case ROTATION:
    /* xa = Ra (x - xa0) + xa0 + ta
       xr0 = xa0 + ta
       x = R(-a) (xa - xa0-ta) + xa0
         = R(-a) (xa - xr0) + xr0 + (xa0-xr0).
    */
    r->type = mappingType[ROTATION];
    r->nPar = 3;
    r->par[2] = -a->par[2];
    r->par[0] = a->x_origin[0] - r->x_origin[0];
    r->par[1] = a->x_origin[1] - r->x_origin[1];
    /* Treat frozen cases:
        - all frozen
        - both translations zero and frozen  => axial rotation.
	- rotations frozen => translation only */
    r->frozen[0] = r->frozen[1] = r->frozen[2] = FALSE;
    if (a->frozen[0] && a->frozen[1]) {     
      if (a->frozen[2]) { /* All frozen */
	r->frozen[0] = r->frozen[1] = r->frozen[2] = TRUE;
      } else if ((abs(a->par[0])<MACHINE_EPS) 
		 && (abs(a->par[1])<MACHINE_EPS)) {
        /* Axial rotation */
	r->frozen[0] = r->frozen[1] = TRUE;
	r->frozen[2] = FALSE;
      }
    } else {
      if (a->frozen[2]) {
	r->frozen[2] = TRUE;
      }
    }
    break;

  case AFFINE:
    r->type = mappingType[AFFINE];
    for(k=0; k<6; k++) r->par[k] = a->par[k];
    r->nPar = 6;
    if (invertAffine(a->par, r->par) != 0) {
      return(-1);
    }
    /* Move center to forward warped center position */
    moveAffineOrigin(a->x_origin, r->par, r->x_origin);
    /* Treat frozen cases:
       - all frozen
       - affine terms frozen
       - axial distortion (translation frozen) */
    for(k=0; k<6; k++) r->frozen[k] = FALSE;
    fz = TRUE;
    for(k=2; k<6; k++)
      if (!a->frozen[k]) {
	fz = FALSE;
	break;
      }
    if (fz) {  /* Affine terms (at least) all frozen */
      for(k=2; k<6; k++) r->frozen[k] = TRUE;
      
      for(k=0; k<2; k++)
	if (!a->frozen[k]) {
	  fz = FALSE;
	  break;
	}
      if (fz)  /* All frozen */
	for(k=0; k<2; k++)  r->frozen[k] = TRUE;

    } else {  /* Affine terms not frozen, check for radial distortion */
      if (a->frozen[0] && a->frozen[1] &&
	  (abs(a->par[0]) < MACHINE_EPS && abs(a->par[1]) < MACHINE_EPS)) {
        /* Axial distortion ok */
	for(k=0; k<2; k++)  r->frozen[k] = TRUE;
      }
    }
    break;
    
  case RATIONAL:
    r->type = mappingType[RATIONAL];
    if (invertRational(a->par, r->par) < 0) 
      return(-1);
    moveRationalOrigin(a->x_origin, r->par, r->x_origin);
    r->nPar = 8;
    /* Treat frozen case:
       - all frozen
    */
    fz = TRUE;
    for(k=0; k<8; k++)
      if (!a->frozen[k]) {
	fz = FALSE;
	break;
      }
    if (fz)  /* all frozen */ 
      for(k=0; k<8; k++) r->frozen[k] = TRUE;
    else
      for(k=0; k<8; k++) r->frozen[k] = FALSE;      
    break;

  default:
    isOk = INVALID_WARP_TYPE;
  } /* end of switch over warp types */

  r->nFree = 0;  
  for(k=0; k<r->nPar; k++)
    if (!r->frozen[k]) r->nFree++;
  return(isOk);
 
}  /** end invertImageWarp */

int convertImageWarpType(ImageWarpStruct *w, int toType)
{
  int k, isOk;
  isOk = 0;
  /* append zeroes to the parameters */
  for(k=w->nPar; k<8; k++) {
    w->par[k] = 0.0;
    w->frozen[k] = TRUE;
  }
  /* Now patch these parameters up in the case they are wrong */
  if (toType == RATIONAL && (getImageWarpType(w)< RATIONAL)) {
    /* Add the identity to the affine coeffs. */
    w->par[2] += 1.0;
    w->par[5] += 1.0;
  }
  switch(toType) {
  case TRANSLATION:
    w->nPar = 2;
    break;
  case ROTATION:
    w->nPar = 3;
    break;
  case AFFINE:
    w->nPar = 6;
    break;
  case RATIONAL:
    w->nPar = 8;
    break;
  default:
    return(INVALID_WARP_TYPE);
  }
  w->nFree=0;
  for(k=0;k<w->nPar;k++)
    if (!w->frozen[k]) w->nFree++;
  return(0);  
}

/* Delete the columns of the image warp jacobian for the frozen parameters */
void compressWarpJac(float jac[2][MAX_MOTION_PAR], int nPar, char *frozen) 
{
  int j;
  int cnt=0;
  for(j=0; j<nPar; j++)
    if (!frozen[j]) {
       jac[0][cnt] = jac[0][j];
       jac[1][cnt] = jac[1][j];
       cnt++;
    }
}

/* Expand a compressed vector of the unfrozen (free) warp parameters
   to the standard form.  Frozen parameters are left as zero in
   this expanded form, which allows for a simple update.
   The result is given in fullPar.  The arrays *freePar and *fullPar
   can be identical storage locations. */
void expandWarpParams(float *freePar, int nPar, int nFree, char *frozen, 
		      float *fullPar)
{
  int j, cnt;

  cnt=nFree-1;
  /* Work backwards through the parameters so that
     the freePar and fullPar can refer to the same array. */
  for(j=nPar-1; j>=0; j--)
    if (!frozen[j]) {
      fullPar[j] = freePar[cnt];
      cnt--;
    } else 
      fullPar[j] = 0.0;
} 


/***************** End of routines on ImageWarpStructs **********************/

/***
 * Do a linear interpolation between low_value and
 * high_value, for position going from 0 to 1, respectively.
 * Returns:
 *  result = low_value * (1 - position) + high_value * position
 ***/
float linear_interp(float low_value, float high_value, float position)
{
  return((float)(low_value * (1.0 - position) + high_value * position));
}

/* ================ Affine 2D Transforms ==================== **
  Definition:
    Affine Transform:
      x_new = affine(x_old; x_origin, a) 
            = x_origin + a_vec + (I + A) (x_old - x_origin)  (Eqn 1a)
            = x_old + a_vec + A (x_old - x_origin)           (Eqn 1b)
    with: 
      x_old - a 2 vector for the pixel coords of the point to be transformed
      x_new - the warped pixel coordinates
      x_origin = the origin for the definition of the affine transform,
               = (x_cent, y_cent)
      a_vec and A the six affine parameters.

    To represent this transform, a float array a[6] is used, with
    the coefficients representing the following values:
      a[6] = (a_vec[0], a_vec[1], A11, A21, A12, A22)

    This is a redundant representation, if you change x_origin then
    an equivalent affine transformation exists with different values
    for a_vec and A (see movAffineOrigin()).

    Note that the image displacement, namely
     d = x_new - x_old,
    is given by:
     d = a_vec + A * (x_old - x_origin)

*================ End: Defn Affine 2D Transform ================*/

/*****
  Affine transform procedures:

  affinePtSample:
    Sample the affine transform 
      x_new = affine(x_old; x_origin, aff)
    with
       float x_origin[2], float aff[6], 
    defined as above.

    Input: 
       x = x_old[0]
       y = x_old[1]

       aff = the 6-d affine parameter vector (a_vec, A) as above.
    On return:
      x_new = affine(x_old; x_origin, aff);
    Returns value: 0 (always) for consistency with other warp sampling
                      routines such as rationalPtSample().
****/
int affinePtSample(float *x_old, float *x_orig, float *aff, float *x_new)
 {
  float loc_x,loc_y;
  /* Local coords */
  loc_x = x_old[0] - x_orig[0];
  loc_y = x_old[1] - x_orig[1];
  /* Compute x_new using equation (Eqn 1b) above */
  x_new[0] = x_old[0] + aff[0] + aff[2] * loc_x + aff[4] * loc_y;
  x_new[1] = x_old[1] + aff[1] + aff[3] * loc_x + aff[5] * loc_y;
  return(0);
 }

/* Return the jacobian of the affine transform with respect to
   the paramters, sampled at image position x_old */
int affineJac(float *x_old, float *x_orig, float *aff, 
	      float jac[2][MAX_MOTION_PAR])
 {
  float loc_x,loc_y;
  int j;

  /* Local coords */
  loc_x = x_old[0] - x_orig[0];
  loc_y = x_old[1] - x_orig[1];

  for(j=0; j<6; j++)
   jac[0][j] = jac[1][j] = 0.0;

  /* Translation terms */
  jac[0][0] = jac[1][1] = 1.0;
  /* See (Eqn 1b) above */
  jac[0][2] = loc_x;
  jac[1][3] = loc_x;
  jac[0][4] = loc_y;
  jac[1][5] = loc_y;

  return(0);
 }

/****
  affineDispSample: 
    Return the image displacement, float d[2], for pixel x_old, 
    according to the affine transform defined by 
       float x_origin[2], float a[6], as above.
    Input: 
       x_diff = x_old - x_origin (a 2-vector)
       aff = the 6-d affine parameter vector (a_vec, A).
    On return:
      displacement d = affine(x_old; x_origin, a) - x_old;
                     = a_vec + A * x_diff   (see Eqn 1b above)
    Return value: 0 (always successful).
****/
int affineDispSample(float *x_diff, float *aff, float *d)
 {
  d[0] = aff[0] + aff[2] * x_diff[0] + aff[4] * x_diff[1] ;
  d[1] = aff[1] + aff[3] * x_diff[0] + aff[5] * x_diff[1] ;
  return(0);
 }

/**** 
  composeAffine:
    Evaluate composition of affine transforms a and b, the
    composite transform is returned in a,
      a <- b o a 
    The origin of the returned composite transform is x_origA
****/
void composeAffine(float *x_origA, float *a, float *x_origB, float *b)
  {
    float scr[6], dx[2];
    int i;

    /* Set dx to be the shift of the origins */
    dx[0] = x_origA[0] - x_origB[0];
    dx[1] = x_origA[1] - x_origB[1];

    /* By Eqn 1b:
         x_new = x_a + b_vec + B (x_a - x_origB)
       where
         x_a = x_old + a_vec + A (x_old - x_origA)
       Expanding the above two equations we get 
         x_new = x_old + [a_vec + b_vec + B(a_vec + dx)]
                 + [A + B + B * A] (x_old - x_origA)
       So:
         a_vec'' = a_vec + b_vec + B(a_vec + dx)
       and 
         A'' = A + B + B * A                           */

    /* a_vec'' = a_vec + b_vec + B(a_vec + dx) */
    scr[0] = a[0] + b[0] + b[2] * (a[0]+dx[0]) + b[4] *(a[1]+dx[1]);
    scr[1] = a[1] + b[1] + b[3] * (a[0]+dx[0]) + b[5] *(a[1]+dx[1]);

    /* A'' = A + B + B * A    */
    scr[2] = a[2] + b[2] + b[2] * a[2] + b[4] * a[3];
    scr[3] = a[3] + b[3] + b[3] * a[2] + b[5] * a[3];
    scr[4] = a[4] + b[4] + b[2] * a[4] + b[4] * a[5];
    scr[5] = a[5] + b[5] + b[3] * a[4] + b[5] * a[5];

    /* Reset a to be a o b */
    for(i=0; i<6; i++)
     a[i] = scr[i];
 
 }

/****
  moveAffineOrigin:
   Move the origin of the affine map from x_origOld to x_origNew
   Coefficients a are updated.
****/
void moveAffineOrigin(float *x_origOld, float *a, float *x_origNew)
{
    float dx[2];

    /* Set dx to be the shift of the origins */
    dx[0] = x_origNew[0] - x_origOld[0];
    dx[1] = x_origNew[1] - x_origOld[1];

    /* By Eqn 1b:
         x_new = x_old + a_vec + A (x_old - x_origOld)
               = x_old + [a_vec + A (x_origNew - x_origOld)]
                       + A (x_old - x_origNew)
               = x_old + [a_vec_new] + A (x_old - x_origNew)
       So  
         a_vec_new = a_vec + A *  dx */
    a[0] += a[2] * dx[0] + a[4] * dx[1];
    a[1] += a[3] * dx[0] + a[5] * dx[1];
}

/****
  invertAffine:
    Compute Inverse Affine Image Transform :
    Affine Transform (Eqn 1a)
       x_new = x_origin + a_vec + (I + A) (x_old - x_origin)
    Inverse Affine Transform:
     x_old = x_orig + b + (I + B) (x_new - x_orig)
    with
     B = inverse(I+A) - I
     b = -(I+B) * a

   Subroutine:
     int invertAffine(a, b)
   
     Input:  Float affine parameters
      a = ( a_0, a_1, A_00, A_10, A_01, A_11)
      (NOTE: Order is down columns first)

     Returns:
      0 if successful
     -1 if failure due to singularity of input transformation.

     Output:  Float affine parameters
      b = ( b_0, b_1, B_00, B_10, B_01, B_11)
     (NOTE: Order is down columns first)
****/
int invertAffine(float *a, float *b)
 {
  float det;

  /* Compute determinant( I+A ) where:
                   | a[2]    a[4] |
      I + A  = I + |              |
                   | a[3]    a[5] |
  */
  det = (1.0 + a[2]) * (1.0 + a[5]) - a[3] * a[4];
  if (abs(det) < MACHINE_EPS) 
   return((int)-1);

  /* set B = inverse(I+A) */
  b[2] = (1.0 + a[5])/det;  /* Switch diagonal elements of (I+A)/det */
  b[5] = (1.0 + a[2])/det;
  b[3] = -a[3]/det;         /* Divide off-diagonal elements by -det */
  b[4] = -a[4]/det;
  /*  now matrix represented by b[2-5] is I+B == inverse(I+A) */
 
  /* Compute b0, b1 according to b = -inverse(I+A) a */
  b[0] = -b[2] * a[0] - b[4] * a[1];
  b[1] = -b[3] * a[0] - b[5] * a[1];
  
  /* Revise b[2-5] to represent B (not I+B, as it does now) */
  b[2] -= 1.0;
  b[5] -= 1.0;
  fprintf(stderr," b= %e %e\n    %e %e %e %e\n", b[0], b[1],
    b[2], b[3], b[4], b[5]);

  return((int) 0);
 }


/* ============ Rational 2D Transforms ============== */

/****
  rationalPtSample:
    Compute sample value, x_new, of rational image transform 
     x_new = x_origin + (a + A(x_vec-x_origin)/(1+b^T (x_vec-x_origin)) (Eqn 2)
    where x_origin = origin for transform.  Here
    x_vec = (x, y)^T, and similarly for x_origin.
   
    Returns:
     ON_HORIZ if the denominator within MACHINE_EPS of 0.0, returns -1
     OUTSIDE_HORIZ if the denominator < -MACHINE_EPS
      0 otherwise (success).
****/
int rationalPtSample(float *x_old, float *x_orig, float *par, float *x_new)
 {
  float scl, dx[2];
  
  dx[0] = x_old[0] - x_orig[0];
  dx[1] = x_old[1] - x_orig[1];
  scl = 1.0 + par[6]*dx[0] + par[7]*dx[1];

  if (abs(scl) <= MACHINE_EPS) {
   /* Invalid argument, denominator apprx= zero */
   x_new[0] = x_new[1] = 0.0;
   return(ON_HORIZ);
  }
  x_new[0] = (par[0] + par[2] * dx[0] + par[4] * dx[1])/scl + x_orig[0];
  x_new[1] = (par[1] + par[3] * dx[0] + par[5] * dx[1])/scl + x_orig[1];

  if (scl < 0.0) {
   /* Invalid argument, denominator < zero */ 
   return(OUTSIDE_HORIZ);
  } else {
   return((int) 0);
  }
 }

/* Return the jacobian of the rational transform, with respect to
   the paramters, at image position x_old */
int rationalJac(float *x_old, float *x_orig, float *par, 
	      float jac[2][MAX_MOTION_PAR])
{
  float dx[2], scl, tmp;
  int i,j;

  for(j=0; j<8; j++)
   jac[0][j] = jac[1][j] = 0.0;

  dx[0] = x_old[0] - x_orig[0];
  dx[1] = x_old[1] - x_orig[1];
  scl = 1.0 + par[6]*dx[0] + par[7]*dx[1];

  if (abs(scl) <= MACHINE_EPS) {
   /* Invalid argument, denominator apprx= zero */
   return(ON_HORIZ);
  } else if (scl < 0.0) {
   /* Invalid argument, denominator < zero */ 
   return(OUTSIDE_HORIZ);
  }

  /**** See (Eqn 2) above, which is equivalent to: 
    x_new[0] = (par[0] + par[2] * dx[0] + par[4] * dx[1])/scl + x_orig[0]
    x_new[1] = (par[1] + par[3] * dx[0] + par[5] * dx[1])/scl + x_orig[1]
  ****/

  /* Translation terms */
  jac[0][0] = jac[1][1] = 1.0/scl;
  /* Affine matrix terms */
  jac[0][2] = dx[0]/scl;
  jac[1][3] = dx[0]/scl;
  jac[0][4] = dx[1]/scl;
  jac[1][5] = dx[1]/scl;
  /* Rational terms */
  tmp = -(par[0] + par[2] * dx[0] + par[4] * dx[1])/(scl*scl);
  jac[0][6] = tmp * dx[0];
  jac[0][7] = tmp * dx[1];
  tmp = -(par[1] + par[3] * dx[0] + par[5] * dx[1])/(scl*scl);
  jac[1][6] = tmp * dx[0];
  jac[1][7] = tmp * dx[1];

  return(0);
}

/***
  invertRational:
    Given rational transform:
      x'-x0 = (a + A(x-x0)/(1+b^T (x-x0))
    where x0 = origin for transform,
    compute inverse
      x-x0 = (a' + A'(x-x0))/(1+b'^T (x-x0))
    
    Here the original transform is given in terms of

    a[0-7] = (a_1 a_2 A_11 A_21 A_12 A_22 b_1 b_2)
    indices.. 0   1   2    3    4    5    6   7

    and similarly, the inverse transform is returned
    in  b[0-7]

    Returns value 0 if successful, else -1 if transform is singular
***/
int invertRational(float *a, float *b)
 {
  float det;

  /* Compute determinant( A ) where:
            | a[2]    a[4] |
      A  =  |              |
            | a[3]    a[5] |
  */
  det = a[2] * a[5] - a[3] * a[4];
  if (abs(det) < MACHINE_EPS) 
   return((int)-1);

  /* set B = inverse(A) */
  b[2] = a[5]/det;  /* Switch diagonal elements of (I+A)/det */
  b[5] = a[2]/det;
  b[3] = -a[3]/det;         /* Divide off-diagonal elements by -det */
  b[4] = -a[4]/det;
  /*  now matrix represented by b[2-5] is B == inverse(A) */
 
  /* Compute b' according to b' = -inverse(A)^T b */
  b[6] = -(b[2] * a[6] + b[3] * a[7]);
  b[7] = -(b[4] * a[6] + b[5] * a[7]);

  /* Compute a' according to a' = -inverse(A) a */
  b[0] = -b[2] * a[0] - b[4] * a[1];
  b[1] = -b[3] * a[0] - b[5] * a[1];
  
  /* Revise b[2-5] to represent 
         A' = inverse(A) - ( b_2, -b_1)^T  (a_2 , -a_1) /det(A)
  */
  b[2] -= a[7] * a[1] /det;
  b[3] -= -a[6] * a[1] /det;
  b[4] -= -a[7] * a[0] /det;
  b[5] -= a[6] * a[0] /det;

  return((int) 0);
 }

/**** 
  composeRational:
    Evaluate composition of affine transforms r and d, the
    composite transform is returned in r,
      r <- d o r 
    The origin of the returned composite transform is x_origR
****/
void composeRational(float *x_origR, float *rFlt,
		     float *x_origD, float *dFlt)
  {
    double r[8], d[8], dOrig[2];
    double scr[8], w, tmp, s[2], Rprime[2][2], Atmp[2][2];
    int i;

    dOrig[0] = x_origR[0] - x_origD[0];
    dOrig[1] = x_origR[1] - x_origD[1];
    for(i=0; i<8; i++) {
     r[i] = (double) rFlt[i];
     d[i] = (double) dFlt[i];
    }

    /* a'' =  a_d + A_d * (a_r+o_r-o_d) - (o_r-o_d){1+b_d^T[a_r+o_r-o_d]} */

    scr[0] = d[0] + d[2] * (r[0]+dOrig[0]) + d[4] * (r[1]+dOrig[1]);
    scr[1] = d[1] + d[3] * (r[0]+dOrig[0]) + d[5] * (r[1]+dOrig[1]);
    tmp = 1 + d[6] * ( r[0] + dOrig[0]) + d[7] * ( r[1] + dOrig[1]);
    scr[0] -= dOrig[0] * tmp;
    scr[1] -= dOrig[1] * tmp;

    /* b'' = b_r^T + b_d dot R'
       with R' = (A_r + (o_r-o_d) b_r^T) */
    Rprime[0][0] = r[2] + dOrig[0] * r[6];
    Rprime[0][1] = r[4] + dOrig[0] * r[7];
    Rprime[1][0] = r[3] + dOrig[1] * r[6];
    Rprime[1][1] = r[5] + dOrig[1] * r[7];

    /* b'' = b_r^T + b_d dot R' */
    scr[6] = r[6] + d[6] * Rprime[0][0] + d[7] * Rprime[1][0] ;
    scr[7] = r[7] + d[6] * Rprime[0][1] + d[7] * Rprime[1][1] ;

    /* A'' = (a_d-(o_r-o_d) ) b_r^T + (A_d - (o_r-o_d) b_d^T) Rprime */

    Atmp[0][0] = d[2] - dOrig[0] * d[6];
    Atmp[0][1] = d[4] - dOrig[0] * d[7];
    Atmp[1][0] = d[3] - dOrig[1] * d[6];
    Atmp[1][1] = d[5] - dOrig[1] * d[7];

    scr[2] = Atmp[0][0] * Rprime[0][0] + Atmp[0][1] * Rprime[1][0];
    scr[3] = Atmp[1][0] * Rprime[0][0] + Atmp[1][1] * Rprime[1][0];
    scr[4] = Atmp[0][0] * Rprime[0][1] + Atmp[0][1] * Rprime[1][1];
    scr[5] = Atmp[1][0] * Rprime[0][1] + Atmp[1][1] * Rprime[1][1];

    scr[2] += (d[0] - dOrig[0]) * r[6];
    scr[4] += (d[0] - dOrig[0]) * r[7];
    scr[3] += (d[1] - dOrig[1]) * r[6];
    scr[5] += (d[1] - dOrig[1]) * r[7];

    /* Rescale by w=1/b0, b0 = new constant term = 1+d_b dot (a_r+o_r-o_d) */

    w = 1.0/(1.0 + d[6] * (r[0]+dOrig[0]) + d[7] * (r[1]+dOrig[1]));
    for(i=0; i<8; i++)
     r[i] = w * scr[i];
 
    for(i=0; i<8; i++) {
     rFlt[i] = (float) r[i];
    }
 }

/* move the origin of the rational map a from x_orig0 to x_orig1 */
void moveRationalOrigin(float *x_orig0, float *a, float *x_orig1)
 {
  float b[8];
  int i;
  /* Make identity map with center x_orig1 */
  for(i=0; i<8; i++)
   b[i] = 0.0;
  b[2] = b[5] = 1.0;
  /* Compose a with identity having x_orig1 as the origin.
  composeRational(x_orig1, b, x_orig0, a);
  /* The composition map b now has the center at x_orig1, as desired. */
  for(i=0; i<8; i++)
   a[i] = b[i];
 }
  
/* Map orientation vector, inVec, at x to new orientation, outVec,
   according to rational mapping having coefficients x_origin, a */
void mapRationalVec(float *inVec, float *x, 
		    float *x_orig, float *a, 
		    float *outVec)
 {
  float tail[2], head[2], tmp[2], nrm;
 
  nrm = inVec[0] * inVec[0] + inVec[1] * inVec[1];
  nrm = sqrt((double) nrm);
  if (nrm = 0.0) {
    outVec[0] = outVec[1] = 0.0;
  } else {
    rationalPtSample(x, x_orig, a, tail);
    tmp[0] = x[0] + inVec[0]/nrm;
    tmp[1] = x[1] + inVec[1]/nrm;
    rationalPtSample(tmp, x_orig, a, head);

    outVec[0] = (head[0] - tail[0]) * nrm;
    outVec[1] = (head[1] - tail[1]) * nrm;
  }  
 }

/* ============ Radial distortion ============== */
/**** 
  radialPtSample:
    Sample the radial distortion map
      x_new = (x_vec - x_origin) * (1 + d0 * ||x_vec - x_origin||^2)
               + x_origin + shift
    That is, given x_vec and the parameters x_origin, shift, d0,
    then find x_new.
    Here array a[3] = (shift_x, shift_y, d0).
    Returns 0 (always successful).
****/
int radialPtSample(float *x, float *x_orig, float *a, float *x_new)
{
   float r, r2, tmp, diff[2];
  
   /* Radius from the origin */
   diff[0] = x[0] - x_orig[0];
   diff[1] = x[1] - x_orig[1];
   r2 = diff[0] * diff[0] + diff[1] * diff[1]; 
   r = sqrt(r2);
   
   tmp = 1.0 + a[2] * r2;
   x_new[0] = x_orig[0] + a[0] + tmp * diff[0];
   x_new[1] = x_orig[1] + a[1] + tmp * diff[1];
   return(0);
}

/**** 
  inverseRadialPtSample:
    Sample the inverse of the radial distortion map
      x_vec = (w - x_origin) * (1 + d0 * ||w - x_origin||^2) + x_origin + shift
    That is, given x_vec and the parameters x_origin, shift, a0,
    then find w.
    Here array a[3] = (shift_x, shift_y, d0).

    Returns 0 if successful, else -1 (failed to solve for inverse point).
****/
int inverseRadialPtSample(float *x, float *x_orig, float *a, float *w)
 {
   float r, r2, rw, res, diff[2];
   int i, maxIts;
  
   maxIts = 10;  /* max number of newton iterations to be used */

   /* Radius from the origin */
   diff[0] = x[0]-a[0]-x_orig[0];
   diff[1] = x[1]-a[1]-x_orig[1];
   r2 = diff[0] * diff[0] + diff[1] * diff[1];
   r = sqrt(r2);
   
   /* radius rw = || w - x_orig || satisfies rw * (1 + a0 *rw^2) - r = 0 */
   if (r < MACHINE_EPS) {
    rw = 0.0;
    w[0] = x_orig[0];
    w[1] = x_orig[1];
   } else { /* Use Newton's method to solve for rw */
    rw = r/(1+ a[0] * r2);
    for(i=0; i< maxIts; i++) {
     res = rw * (1+ a[2] * rw * rw) - r;
     if (abs(res)/r <= MACHINE_EPS) 
      break;
     rw -= res/( 1 + 3.0 * a[2] * rw * rw); /* Newton step */
    }

    if (i>=maxIts) {
     fprintf(stderr, "InverseWarp: x,y = %e %e takes more than %d its\n",
              x[0],x[1],i );
     return(-1);
    }

    /* Compute corrected point, which is at a distance rw in the direction
       diff[*]/r away from x_orig.  */
    w[0] = x_orig[0] + rw/r * diff[0];
    w[1] = x_orig[1] + rw/r * diff[1];
   }
   return(0);
 }

/* ============ Image Warping ============== */

/*
 Backwards warp of a nx by ny image by an affine warp with parameters
 given by (x_origin, b).  
 That is, for pixel (x,y):
   warpedImage(x,y) =(apprx) origImage(affine(x,y; x_origin, b))
 where affine(x,y; x_origin, b) gives the pixel coords for
 the (backwards) affine warp evaluated at (x,y).

 The result is stored in the float array warped, which must be
 already allocated to hold at least an nx * ny array.

 Pixel values are determined by linear interpolation from the original image.

 While this is fast and simple, a better interpolation
 may be desirable.

 WARNING: Filtering would be essential if the image is to be
 subsampled significantly.
*/
void affine_warp_image(float *image, int nx, int ny,
		       float *warped, 
		       float *x_orig, float *b, 
		       unsigned char defaultGrey)
{
  float xCurrent[2], x[2], rx[2];
  int ixy, ix, iy, floor_x, floor_y, ceil_x, ceil_y;

  for(iy=0;iy<ny;iy++){
    xCurrent[1] = (float) iy;
    for(ix=0;ix<nx;ix++){
      ixy = (iy*nx) + ix;
      xCurrent[0] = (float) ix;
      affinePtSample(xCurrent, x_orig, b, x);

      if (x[0] < 0 || x[0] > (nx - 1) || x[1] < 0 || x[1] > (ny - 1)) 
	warped[ixy] = defaultGrey;
      else{
	floor_x = ((int) floor(x[0]));
	ceil_x = ((int) ceil(x[0]));
	rx[0] = x[0] - ((float) floor_x);

	floor_y = ((int) floor(x[1]));
	ceil_y = ((int) ceil(x[1]));
	rx[1] = x[1] - ((float) floor_y);
      
	warped[ixy] = 
	  linear_interp
	  (linear_interp(image[(floor_y*nx)+floor_x],
			 image[(ceil_y*nx)+floor_x],
			 rx[1]),
	   linear_interp(image[(floor_y*nx)+ceil_x],
			 image[(ceil_y*nx)+ceil_x],
			 rx[1]),
	   rx[0]);
      }
    }
  }
}

/*
Backwards warp of "image" by warp b.  The new image is given by
 temp(x,y) = image(affine(x,y; x_origin, b))
This image is sampled at pixels:
 (x,y) = (xMin, yMin) + (dx * i, dy * j) 
to form the float image warped(i,j).

This is similar to affine_warp_image() above, but
here the warped image can be resampled at a different rate.

Pixel values are determined by linearly interpolating the warped image.
No filtering is done.  This is needed for some warps in order to control
aliasing. */
void affine_warp_sample_image(float *image, int nx, int ny,
                  float *warped, 
                  float xMin, float yMin, float dx, float dy, 
                  int newNx, int newNy,
                  float *x_orig, float *b, 
		  unsigned char defaultGrey)
{
  float xOld[2], x[2], rx[2];
  int ixy, ix, iy, floor_x, floor_y, ceil_x, ceil_y;

  for(iy=0;iy<newNy;iy++){
    xOld[1] = yMin + dy*iy;
    for(ix=0;ix<newNx;ix++){
      xOld[0] = xMin + dx*ix;
      ixy = (iy*nx) + ix;
      affinePtSample(xOld, x_orig, b, x);

      if (x[0] < 0 || x[0] > (nx - 1) || x[1] < 0 || x[1] > (ny - 1)) 
	warped[ixy] = defaultGrey;
      else{
	floor_x = floor(x[0]);
	ceil_x = ceil(x[0]);
	rx[0] = x[0] - ((float) floor_x);

	floor_y = ((int) floor(x[1]));
	ceil_y = ((int) ceil(x[1]));
	rx[1] = x[1] - ((float) floor_y);
      
	warped[ixy] = 
	  linear_interp
	  (linear_interp(image[(floor_y*nx)+floor_x],
			 image[(ceil_y*nx)+floor_x],
			 rx[1]),
	   linear_interp(image[(floor_y*nx)+ceil_x],
			 image[(ceil_y*nx)+ceil_x],
			 rx[1]),
	   rx[0]);
      }
    }
  }
}

/*
Backwards warp of "image" by rational warp b.
The result is stored in "warped".
Pixel values are determined by linearly interpolating the warped image.
While this is fairly fast, a better interpolation may be desirable.
*/
void rational_warp_image(float *image, int nx, int ny,
			 float *warped, 
			 float *x_orig, float *b,
			 unsigned char defaultGrey)
{
  float xCurrent[2], x[2], rx[2];
  int ixy, ix, iy, floor_x, floor_y, ceil_x, ceil_y;

  for(iy=0;iy<ny;iy++){
    xCurrent[1] = (float) iy;
    for(ix=0;ix<nx;ix++){
      ixy = (iy*nx) + ix;
      xCurrent[0] = (float) ix;
      if (rationalPtSample(xCurrent, x_orig, b, x) <0 ) {
	/* xCurrent is out of range for the rational map */
        x[0] = -1.0; /* Forces the use of defaultGrey below */
      }

      if (x[0] < 0 || x[0] > (nx - 1) || x[1] < 0 || x[1] > (ny - 1)) 
	warped[ixy] = defaultGrey;
      else{
	floor_x = ((int) floor(x[0]));
	ceil_x = ((int) ceil(x[0]));
	rx[0] = x[0] - ((float) floor_x);

	floor_y = ((int) floor(x[1]));
	ceil_y = ((int) ceil(x[1]));
	rx[1] = x[1] - ((float) floor_y);
      
	warped[ixy] = 
	  linear_interp
	  (linear_interp(image[(floor_y*nx)+floor_x],
			 image[(ceil_y*nx)+floor_x],
			 rx[1]),
	   linear_interp(image[(floor_y*nx)+ceil_x],
			 image[(ceil_y*nx)+ceil_x],
			 rx[1]),
	   rx[0]);
      }
    }
  }
}

/*
Backwards warp of "image" according to radial warp.
The result is stored in "warped".
Pixel values are determined by linearly interpolating the warped image.
While this is fairly fast, a better interpolation may be desirable.
*/
void radial_warp_image(float *image, int nx, int ny,
		       float *warped, 
		       float *x_orig, float *par, 
		       unsigned char defaultGrey)
{
  float xCurrent[2], x[2], rx[2];
  int ixy, ix, iy, floor_x, floor_y, ceil_x, ceil_y;

  for(iy=0;iy<ny;iy++){
    xCurrent[1] = (float) iy;
    for(ix=0;ix<nx;ix++){
      ixy = (iy*nx) + ix;
      xCurrent[0] = (float) ix;
      if (inverseRadialPtSample(xCurrent, x_orig, par, x) < 0) {
	x[0] = -1.0; /* Invalid inverse warp, force the pixel to be
                        set to defaultGrey below */
      }

      if (x[0] < 0 || x[0] > (nx - 1) || x[1] < 0 || x[1] > (ny - 1)) 
	warped[ixy] = defaultGrey;
      else{
	floor_x = ((int) floor(x[0]));
	ceil_x = ((int) ceil(x[0]));
	rx[0] = x[0] - ((float) floor_x);

	floor_y = ((int) floor(x[1]));
	ceil_y = ((int) ceil(x[1]));
	rx[1] = x[1] - ((float) floor_y);
      
	warped[ixy] = 
	  linear_interp
	  (linear_interp(image[(floor_y*nx)+floor_x],
			 image[(ceil_y*nx)+floor_x],
			 rx[1]),
	   linear_interp(image[(floor_y*nx)+ceil_x],
			 image[(ceil_y*nx)+ceil_x],
			 rx[1]),
	   rx[0]);
      }
    }
  }
}

/****
 warpAndBound:
  Forwards warp of "image" by rational warp a.
  The result is stored in "warped".
  The size of warped is the smallest upright rectangle containing
  the mapping of "image".  The bounding box is restricted to lie on
  a subsampled grid, starting at 0,0 with spacing `subSample'.
  The size of warped is returned as *pnxw, *pnyw.
  Grey levels are determined by linearly interpolating the warped image.
  While this is fairly fast, a better interpolation may be desirable.

  Returns the backwards warp from new warped image coords to original
  image in b with center x_origin left unchanged.  That is:
     warped(x,y) = image(rationalWarp(x,y; x_orig, b))
****/
void warpAndBound(float *image, int nx, int ny, 
		  float **pwarped, int *pnxw, int *pnyw, 
		  float *x_orig, float *a, 
		  unsigned char defaultGrey,
		  int subSample, float *b)
{
  float *warped;
  float c[8], boxCorners[2][2], imageVertex[4][2];

  float xCurrent[2], x[2], rx[2];
  int ixy, ix, iy, floor_x, floor_y, ceil_x, ceil_y;
  int i,j, ix0, iy0, ix1, iy1, size;

  /* compute inverse mapping (check for singularity) */
  if (-1 == invertRational(a, c))
    error("Rational transform for prewarp nearly singular","");
  /* c maps from warped coords to original "image" coords */

  /* Computed bounding upright rectangle */

  /* Compute box corners */
  xCurrent[0] = 0.0; xCurrent[1] = 0.0;
  rationalPtSample(xCurrent, x_orig, a, imageVertex[0]);
  xCurrent[1] = (float)(ny-1);
  rationalPtSample(xCurrent, x_orig, a, imageVertex[1]);
  xCurrent[0] = (float)(nx-1);
  rationalPtSample(xCurrent, x_orig, a, imageVertex[2]);
  xCurrent[1] = 0.0;
  rationalPtSample(xCurrent, x_orig, a, imageVertex[3]);

  boxCorners[0][0] = imageVertex[0][0];
  for(j=1; j<4; j++)
        boxCorners[0][0] = min(boxCorners[0][0],  imageVertex[j][0]);
  boxCorners[0][1] = imageVertex[0][1];
  for(j=1; j<4; j++)
        boxCorners[0][1] = min(boxCorners[0][1],  imageVertex[j][1]);
  boxCorners[1][0] = imageVertex[0][0];
  for(j=1; j<4; j++)
        boxCorners[1][0] = max(boxCorners[1][0],  imageVertex[j][0]);
  boxCorners[1][1] = imageVertex[0][1];
  for(j=1; j<4; j++)
        boxCorners[1][1] = max(boxCorners[1][1],  imageVertex[j][1]);
  /* Choose box placement to match subSampled grid */
  if (subSample <= 0) subSample = 1; 
  ix0 = floor(boxCorners[0][0]/subSample) * subSample;
  iy0 = floor(boxCorners[0][1]/subSample) * subSample;
  ix1 = ceil(boxCorners[1][0]/subSample) * subSample;
  iy1 = ceil(boxCorners[1][1]/subSample) * subSample;

  *pnxw = ix1 - ix0;
  *pnyw = iy1 - iy0;
  size = (*pnxw) * (*pnyw) ;
  fprintf(stderr," Warped template size %d %d\n", *pnxw, *pnyw);
  grabFloatMemory(pwarped, size, "warped image");
  warped = *pwarped;

  /* Modify backwards warp c according to the translation
     of coords which takes top left corner of warped image,
     in canonical coords, namely (0, 0), to pixel position
     (ix0, iy0) */
  for(i=0; i<8; i++)
   b[i] = 0.0;
  b[2] = b[5] = 1.0;
  b[0] = (float) ix0;
  b[1] = (float) iy0;
  /* b <= c o b */
  composeRational(x_orig, b,  x_orig, c);

  for(iy=0;iy<(*pnyw);iy++){
    xCurrent[1] = (float) iy;
    for(ix=0;ix<(*pnxw);ix++){
      ixy = iy * (*pnxw) + ix;
      xCurrent[0] = (float) ix;
      /* Map backwards from warp pixel ix,iy, to orig image pixel x[*] */
      if (rationalPtSample(xCurrent, x_orig, b, x) <0 ) {
	/* xCurrent is out of range for the rational map */
        x[0] = -1.0; /* Forces the use of defaultGrey below */
      }

      if (x[0] < 0 || x[0] > (nx - 1) || x[1] < 0 || x[1] > (ny - 1)) 
	warped[ixy] = defaultGrey;
      else{
	floor_x = ((int) floor(x[0]));
	ceil_x = ((int) ceil(x[0]));
	rx[0] = x[0] - ((float) floor_x);

	floor_y = ((int) floor(x[1]));
	ceil_y = ((int) ceil(x[1]));
	rx[1] = x[1] - ((float) floor_y);
      
	warped[ixy] = 
	  linear_interp
	  (linear_interp(image[(floor_y*nx)+floor_x],
			 image[(ceil_y*nx)+floor_x],
			 rx[1]),
	   linear_interp(image[(floor_y*nx)+ceil_x],
			 image[(ceil_y*nx)+ceil_x],
			 rx[1]),
	   rx[0]);
      }
    }
  }
}

/****
 warpCrop:
  Forwards warp of "image" by rational warp a.
  The result is stored in "warped".

  The size of warped cropped to bounding box with top left
  corner ix0,iy0, and bottom right corner ixy, iy1.
  Grey levels are determined by linearly interpolating the warped image.
  While this is fairly fast, a better interpolation may be desirable.
****/
void warpCrop(float *image, int nx, int ny, 
	      float **pwarped, 
              int ix0, int iy0, int ix1, int iy1, 
	      float *x_orig, float *a,
              unsigned char defaultGrey)
{
  float *warped;
  float b[8], c[8], xCurrent[2], x[2], rx[2];
  int i,j, ixy, ix, iy, floor_x, floor_y, ceil_x, ceil_y;
  int nxw, nyw, size;

  /* compute inverse mapping (check for singularity) */
  if (-1 == invertRational(a, c))
    error("Rational transform for prewarp nearly singular","");
  /* c maps from warped coords to original "image" coords */

  /* Computed bounding upright rectangle */

  nxw = ix1 - ix0;
  nyw = iy1 - iy0;
  size = nxw * nyw;
  fprintf(stderr," Warped template size %d %d\n", nxw, nyw);
  grabFloatMemory(pwarped, size, "warped image");
  warped = *pwarped;

  /* Modify backwards warp c according to the translation
     of coords which takes top left corner of warped image,
     in canonical coords, namely (0, 0), to pixel position
     (ix0, iy0) */
  for(i=0; i<8; i++)
   b[i] = 0.0;
  b[2] = b[5] = 1.0;
  b[0] = (float) ix0;
  b[1] = (float) iy0;
  /* b <= c o b */
  composeRational(x_orig, b,  x_orig, c);

  fprintf(stderr,"warpCrop params:\n");
  fprintf(stderr,"%6.1f %6.1f\n", x_orig[0], x_orig[1]);
  fprintf(stderr,"%e %e\n", b[0], b[1]);
  fprintf(stderr,"%e %e %e %e\n", b[2], b[3], b[4], b[5]);
  fprintf(stderr,"%e %e\n", b[6], b[7]);

  for(iy=0;iy<nyw;iy++){
    xCurrent[1] = (float) iy;
    for(ix=0;ix<nxw;ix++){
      ixy = iy * nxw + ix;
      xCurrent[0] = (float) ix;
      /* Map backwards from warp pixel ix,iy, to orig image pixel x[*] */
      if (rationalPtSample(xCurrent, x_orig, b, x) <0 ) {
	/* xCurrent is out of range for the rational map */
        x[0] = -1.0; /* Forces the use of defaultGrey below */
      }

      if (x[0] < 0 || x[0] > (nx - 1) || x[1] < 0 || x[1] > (ny - 1)) 
	warped[ixy] = defaultGrey;
      else{
	floor_x = ((int) floor(x[0]));
	ceil_x = ((int) ceil(x[0]));
	rx[0] = x[0] - ((float) floor_x);

	floor_y = ((int) floor(x[1]));
	ceil_y = ((int) ceil(x[1]));
	rx[1] = x[1] - ((float) floor_y);
      
	warped[ixy] = 
	  linear_interp
	  (linear_interp(image[(floor_y*nx)+floor_x],
			 image[(ceil_y*nx)+floor_x],
			 rx[1]),
	   linear_interp(image[(floor_y*nx)+ceil_x],
			 image[(ceil_y*nx)+ceil_x],
			 rx[1]),
	   rx[0]);
      }
    }
  }
}
